@@ -3,25 +3,37 @@
 # The main purpose of this class is to easily reload a training 
 #---------------------------------------------------------------------------
 
-import os
-from skimage.io import imsave
+import os 
+import shutil
 import torch
 from torch.utils.data import DataLoader
 # from monai.data import ThreadDataLoader
-import pandas as pd
 # import torch.distributed as dist
+import numpy as np
 
 from biom3d import register
 from biom3d import callbacks as clbk
 from biom3d import utils
+from biom3d.utils import adaptive_imsave
 
 #---------------------------------------------------------------------------
 # utils to read config's functions in the function register
 
 def read_config(config_fct, register_cat, **kwargs):
-    """
-    read the register category at register_cat and run the corresponding function 
-    named in the config file
+    """Read the config function in the register category and run the corresponding function with the keyword arguments which are merged from 1. the register kwargs, 2. the config file kwargs and 3. this function kwargs.
+
+    Parameters
+    ----------
+    config_fct : str
+        Name of the function listed in the register.
+    register_cat : Dict
+        Dictionary defining one category in the register.
+    **kwargs : dict, optional
+        Additional keyword arguments of the function defined by the config_fct
+        
+    Returns
+    -------
+    The eventual outputs of the function.
     """
     # read the corresponding name in the register
     register_fct_ = register_cat[config_fct.fct]
@@ -111,46 +123,105 @@ class LARS(torch.optim.Optimizer):
 # Global routine
 
 class Builder:
-    """
-    the original goal of the builder was to easily save and load a model training.
+    """The Builder is the core of biom3d. 
+
+    The Builder reads a configuration file and builds the components required for training and prediction.
+
+    Please note that in this current version the training and inference can only be done with CUDA.
+
+    Multi-GPUs training is also supported with DataParallel only (Distributed Data Parallel is not).
+
+    Training is currently done with the SGD optimizer. If you would like to change the optimizer, you can edit `self.build_training` method.
+
+    If both `config` and `path` are defined then Builder considers that fine-tuning is intended.
+
+    If `path` is a list of path, then multi-model prediction will be used, training should be off/False.
+
+    Parameters
+    ----------
+    config : str, dict or biom3d.utils.Dict
+        Path to a Python configuration file (in either .py or .yaml format) or dictionary of a configuration file. Please refer to biom3d.config_default.py to see the default configuration file format.
+    path : str, list of str
+        Path to a builder folder which contains the model folder, the model configuration and the training logs.
+        If path is a list of strings, then it is considered that it is intended to run multi-model predictions. Training is not compatible with this mode.
+    training : bool, default=True
+        Whether to load the model in training or testing mode.
+
+    Examples
+    --------
+    To run a training from a configuration file do:
+
+    >>> from importlib import import_module
+    >>> cfg = import_module("biom3d.configs.unet_default").CONFIG
+    >>> builder = Builder(config=cfg)
+    >>> builder.run_training() # start the training
+
+    To run a prediction from a log folder, do:
+    
+    >>> path = "path/to/log/folder"
+    >>> builder = Builder(path=path, training=False)
+    >>> builder.run_prediction_folder(dir_in="input/folder", dir_out="output/folder")
     """
     def __init__(self, 
         config=None,         # inherit from Config class, stores the global variables
         path=None,      # path to a training folder
         training=True,  # use training mode or testing?
         ):                
+        # for training or fine-tuning:
+        # load the config file and change some parameters if multi-gpus training
+        if config is not None: 
+            assert type(config)==str or type(config)==dict or type(config)==utils.Dict, "[Error] Config has the wrong type {}".format(type(config))
+            if type(config)==str:
+                self.config_path = config
+                self.config = utils.adaptive_load_config(config)
+            else:
+                self.config_path = None
+                self.config = config
 
-        if path is not None:
-            self.config = utils.load_config(path + "/log/config.yaml")
+            # if there are more than 1 GPU we augment the batch size and reduce the number of epochs
+            if torch.cuda.device_count() > 1: 
+                print("Let's use", torch.cuda.device_count(), "GPUs!")
+
+                # Loop through all key-value pairs of a nested dictionary and change the batch_size 
+                self.config = utils.nested_dict_change_value(self.config, 'batch_size', torch.cuda.device_count()*self.config.BATCH_SIZE)
+                self.config.BATCH_SIZE *= torch.cuda.device_count()
+                self.config.NB_EPOCHS = self.config.NB_EPOCHS//torch.cuda.device_count()
+
+        # fine-tuning  
+        if path is not None and config is not None:
+            print("Fine-tuning mode! The path to a builder folder and a configuration file have been input.")
+            
+            # build the training folder
+            self.build_train()
+
+            # load the model weights
+            model_dir = os.path.join(path, 'model')
+            model_name = utils.load_yaml_config(os.path.join(path,"log","config.yaml")).DESC+'.pth'
+            ckpt_path = os.path.join(model_dir, model_name)
+            ckpt = torch.load(ckpt_path)
+            print("Loading model from", ckpt_path)
+            print(self.model.load_state_dict(ckpt['model'], strict=False))
+        
+        # training restart or prediction
+        elif path is not None:
             # print(self.config)
             if training:
                 self.load_train(path)
             else:
                 self.load_test(path)
+        
+        # standard training
         else:
             assert config is not None, "[Error] config file not defined."
-            self.config = config
-
-            # if there are more than 1 GPU we augment the batch size and reduce the number of epochs
-            if torch.cuda.device_count() > 1: 
-                print("Let's use", torch.cuda.device_count(), "GPUs!")
-                # Loop through all key-value pairs of a nested dictionary and change the batch_size 
-                # for pairs in utils.nested_dict_pairs_iterator(self.config):
-                #     if 'batch_size' in pairs:
-                #         save = self.config[pairs[0]]; i=1
-                #         while i < len(pairs) and pairs[i]!='batch_size':
-                #             save = save[pairs[i]]; i+=1
-                #         save['batch_size'] = torch.cuda.device_count()*self.config.BATCH_SIZE
-                self.config = utils.nested_dict_change_value(self.config, 'batch_size', torch.cuda.device_count()*self.config.BATCH_SIZE)
-                self.config.BATCH_SIZE *= torch.cuda.device_count()
-                self.config.NB_EPOCHS = self.config.NB_EPOCHS//torch.cuda.device_count()
-
             # print(self.config)
             self.build_train()
+        
+        # if cuda is not available then deactivate USE_FP16
+        if not torch.cuda.is_available():
+            self.config.USE_FP16 = False
 
     def build_dataset(self):
-        """
-        build the dataset.
+        """Build the dataset.
         """
         if 'TRAIN_DATASET' in self.config.keys():
             self.train_dataset = read_config(self.config.TRAIN_DATASET, register.datasets)
@@ -168,8 +239,7 @@ class Builder:
 
 
     def build_model(self, training=True):
-        """
-        build the model, the losses and the metrics.
+        """Build the model, the losses and the metrics.
         """
         
         # self-supervised models are special cases
@@ -203,13 +273,11 @@ class Builder:
         else: 
             self.model = read_config(self.config.MODEL, register.models)
 
-
-            if torch.cuda.device_count() > 1:
-                print("Let's use", torch.cuda.device_count(), "GPUs!")
-                self.model = torch.nn.DataParallel(self.model)
-
             if torch.cuda.is_available():
                 self.model.cuda()
+                if torch.cuda.device_count() > 1:
+                    print("Let's use", torch.cuda.device_count(), "GPUs!")
+                    self.model = torch.nn.DataParallel(self.model)
 
             # TODO: use DDP...
             # if rank is not None: 
@@ -225,15 +293,16 @@ class Builder:
 
         # losses
         if training:
-            # self.loss_fn = DiceBCE(name='train_loss') # normal supervision
             self.loss_fn = read_config(self.config.TRAIN_LOSS, register.metrics)
-            self.loss_fn.cuda().train()
-            # self.loss_fn.train()
+            if torch.cuda.is_available():
+                self.loss_fn.cuda()
+            self.loss_fn.train()
 
             if 'VAL_LOSS' in self.config.keys():
                 self.val_loss_fn = read_config(self.config.VAL_LOSS, register.metrics)
-                self.val_loss_fn.cuda().eval()
-                # self.val_loss_fn.eval()
+                if torch.cuda.is_available():
+                    self.val_loss_fn.cuda()
+                self.val_loss_fn.eval()
             else: 
                 self.val_loss_fn = None
 
@@ -241,12 +310,18 @@ class Builder:
             if 'TRAIN_METRICS' in self.config.keys():
                 self.train_metrics = [read_config(v, register.metrics) for v in self.config.TRAIN_METRICS.values()]
                 # print(self.train_metrics)
-                for m in self.train_metrics: m.cuda().eval()
+                for m in self.train_metrics: 
+                    if torch.cuda.is_available():
+                        m.cuda()
+                    m.eval()
             else: self.train_metrics = []
 
             if 'VAL_METRICS' in self.config.keys():
                 self.val_metrics = [read_config(v, register.metrics) for v in self.config.VAL_METRICS.values()]
-                for m in self.val_metrics: m.cuda().eval()
+                for m in self.val_metrics: 
+                    if torch.cuda.is_available():
+                        m.cuda()
+                    m.eval()
             else: self.val_metrics = []
 
             # optimizer
@@ -264,14 +339,13 @@ class Builder:
                 # [{'params': self.model.parameters()}, {'params': self.loss_fn.parameters()}], 
                 params, 
                 lr=lr, momentum=0.99, nesterov=True, weight_decay=weight_decay)
-                # lr=lr, momentum=0.99, nesterov=True)
-
             # self.optim = LARS(params)
 
 
     def build_callbacks(self):
-        """
-        build the callbacks for the training process.
+        """Build the callbacks for the training process. Callback are used to monitor the training process and to update the schedulers.
+
+        As the callbacks are often dependant on the Builder arguments, they are defined directly here.
         """
 
         clbk_dict = {}
@@ -368,15 +442,16 @@ class Builder:
                 saved_loss=self.loss_fn)
         clbk_dict["model_saver"] = self.clbk_modelsaver
 
-        # self.clbk_logsaver = clbk.LogSaver(
-        #         log_dir=self.log_dir,
-        #         train_loss=self.loss_fn,
-        #         val_loss=self.val_loss_fn,
-        #         train_metrics=self.train_metrics,
-        #         val_metrics=self.val_metrics,
-        #         scheduler=self.clbk_scheduler,
-        #         save_best=self.config.SAVE_BEST,
-        #         every_batch=10)
+        self.clbk_logsaver = clbk.LogSaver(
+                log_dir=self.log_dir,
+                train_loss=self.loss_fn,
+                val_loss=None if not hasattr(self, 'val_loss_fn') else self.val_loss_fn,
+                train_metrics=None if not hasattr(self, 'train_metrics') else self.train_metrics,
+                val_metrics=None if not hasattr(self, 'val_metrics') else self.val_metrics,
+                scheduler=None if not hasattr(self, 'clbk_scheduler') else self.clbk_scheduler,)
+                # save_best=self.config.SAVE_BEST,
+                # every_batch=10)
+        clbk_dict["log_saver"] = self.clbk_logsaver
         
         if "USE_IMAGE_CLBK" in self.config.keys() and self.config.USE_IMAGE_CLBK and hasattr(self, 'val_dataloader'):
             self.clbk_imagesaver = clbk.ImageSaver(
@@ -388,7 +463,11 @@ class Builder:
                     )
             clbk_dict["image_saver"] = self.clbk_imagesaver
 
-        
+        self.clbk_metricupdater = clbk.MetricsUpdater(
+        metrics=[self.loss_fn] + self.train_metrics if self.train_metrics else [self.loss_fn], 
+        batch_size=self.config.BATCH_SIZE)
+        clbk_dict["metric_updater"] = self.clbk_metricupdater
+
         self.clbk_tensorboardsaver = clbk.TensorboardSaver(
                 log_dir=self.log_dir,
                 train_loss=self.loss_fn,
@@ -411,23 +490,40 @@ class Builder:
         #         loss=self.loss_fn,
         #         test_loss=self.val_loss_fn)
         # clbk_list += [self.clbk_telegram]
-        
-        self.clbk_metricupdater = clbk.MetricsUpdater(
-                metrics=[self.loss_fn] + self.train_metrics if self.train_metrics else [self.loss_fn], 
-                batch_size=self.config.BATCH_SIZE)
-        clbk_dict["metric_updater"] = self.clbk_metricupdater
 
         self.callbacks = clbk.Callbacks(clbk_dict)
 
 
     def build_train(self):
+        """Successively build the dataset, the model and the callbacks.
         """
-        build the method from scratch.
-        """
-        # saver configs
+        # make it deterministic 
+        np.random.seed(12345)
+        torch.manual_seed(12345)
+        if torch.cuda.is_available():
+            torch.cuda.manual_seed_all(12345)
+        torch.backends.cudnn.deterministic = True
+        torch.backends.cudnn.benchmark = False
+
+        # saver folder configuration
+        folder_name = self.config.DESC
+        
+        # add fold number if defined
+        if 'FOLD' in self.config.keys():
+            folder_name += '_fold'+str(self.config.FOLD)
         self.base_dir, self.image_dir, self.log_dir, self.model_dir = utils.create_save_dirs(
-            self.config.LOG_DIR, self.config.DESC, dir_names=['image', 'log', 'model'], return_base_dir=True) 
-        utils.save_config(os.path.join(self.log_dir, 'config.yaml'), self.config) # save the config file
+            self.config.LOG_DIR, folder_name, dir_names=['image', 'log', 'model'], return_base_dir=True) 
+    
+        # save the config file
+        if self.config_path is not None:
+            basename = os.path.basename(self.config_path)
+            shutil.copy(self.config_path, os.path.join(self.log_dir, basename))
+        # copy csv file
+        if self.config.CSV_DIR is not None:
+            basename = os.path.basename(self.config.CSV_DIR)
+            shutil.copy(self.config.CSV_DIR, os.path.join(self.log_dir, basename))
+
+        utils.save_yaml_config(os.path.join(self.log_dir, 'config.yaml'), self.config) # will eventually replace the yaml file
 
         self.model_path = os.path.join(self.model_dir, self.config.DESC)
 
@@ -440,8 +536,7 @@ class Builder:
         self.build_callbacks()
 
     def run_training(self):
-        """
-        run the training and validation routines.
+        """Run the training and validation routines.
         """
         if 'USE_FP16' in self.config.keys() and self.config.USE_FP16:
             scaler = torch.cuda.amp.GradScaler()
@@ -472,7 +567,8 @@ class Builder:
                         model=self.model,
                         loss_fn=self.val_loss_fn,
                         metrics=self.val_metrics,
-                        use_fp16=scaler is not None)
+                        use_fp16=scaler is not None,
+                        use_deep_supervision=self.config.USE_DEEP_SUPERVISION,)
             self.callbacks.on_epoch_end(epoch)
         self.callbacks.on_train_end()
     
@@ -493,48 +589,133 @@ class Builder:
     #     )
 
     def run_prediction_single(self, img_path, return_logit=True):
-        """
-        TODO: this function should be defined in another file 
-        compute a prediction for one image, just for visualizing
-        """
+        """Compute a prediction for one image using the predictor defined in the configuration file.
 
-        return read_config(
-            self.config.PREDICTOR, 
-            register.predictors,
-            img_path = img_path,
-            model = self.model,
-            return_logit = return_logit,
-            )
-    
-    def run_prediction_folder(self, dir_in, dir_out, return_logit=False):
+        Parameters
+        ----------
+        img_path : str
+            Path to the image.
+        return_logit : bool, default=True
+            Whether to return the logit, i.e. the model output before the final activation. 
+        
+        Returns
+        -------
+        numpy.ndarray
+            Output images.
         """
-        compute a prediction for one image, just for visualizing
+        # load image with the preprocessor
+        # img, img_metadata = read_config(self.config.PREPROCESSOR, register.preprocessors, img_path=img_path)
+        if type(self.config)==list: # multi-model mode!
+            # check if the preprocessing are all equal, then only use one preprocessing
+            # TODO: make it more flexible?
+            assert np.all([config.PREPROCESSOR==self.config[0].PREPROCESSOR for config in self.config[1:]]), "[Error] For multi-model prediction, the current version of biom3d imposes that all preprocessor are identical."
+            
+            # preprocessing
+            img, img_meta = read_config(self.config[0].PREPROCESSOR, register.preprocessors, img_path=img_path)
+
+            # same for postprocessors
+            for i in range(len(self.config)):
+                if not 'POSTPROCESSOR' in self.config[i].keys():
+                    self.config[i].POSTPROCESSOR = utils.Dict(fct="Seg", kwargs=utils.Dict())
+
+            assert np.all([config.POSTPROCESSOR==self.config[0].POSTPROCESSOR for config in self.config[1:]]), "[Error] For multi-model prediction, the current version of biom3d imposes that all postprocessors are identical."
+
+            logit = None # to accumulate the logit
+            for i, config in enumerate(self.config):
+                # prediction
+                print('Running prediction for model number', i)
+                out = read_config(
+                    config.PREDICTOR, 
+                    register.predictors,
+                    img = img,
+                    model = self.model[i], # prediction for model i
+                    **img_meta
+                    )
+
+                # accumulate the logit
+                logit = out if logit is None else logit+out
+
+            logit /= len(self.config)
+
+            # final post-processing
+            return read_config(
+                    self.config[0].POSTPROCESSOR, 
+                    register.postprocessors,
+                    logit = logit, 
+                    return_logit = return_logit,
+                    **img_meta) # all img_meta should be equal as we use the same preprocessors
+        else:
+            img, img_meta = read_config(self.config.PREPROCESSOR, register.preprocessors, img_path=img_path)
+
+            # same for postprocessors
+            if not 'POSTPROCESSOR' in self.config.keys():
+                self.config.POSTPROCESSOR = utils.Dict(fct="Seg", kwargs=utils.Dict())
+
+            # prediction
+            out = read_config(
+                self.config.PREDICTOR, 
+                register.predictors,
+                img = img,
+                model = self.model,
+                **img_meta)
+            
+            # postprocessing
+            return read_config(
+                self.config.POSTPROCESSOR, 
+                register.postprocessors,
+                logit = out,
+                return_logit = return_logit,
+                **img_meta)
+
+    def run_prediction_folder(self, dir_in, dir_out, return_logit=False):
+        """Compute predictions for a folder of images.
+
+        Parameters
+        ----------
+        dir_in : str
+            Path to the input folder of images.
+        dir_out : str
+            Path to the output folder where the predictions will be stored.
+        return_logit : bool, default=False
+            Whether to save the logit, i.e. the model output before the final activation.
         """
         fnames_in = sorted(utils.abs_listdir(dir_in))
         if not os.path.isdir(dir_out):
             os.makedirs(dir_out, exist_ok=True)
         
         # remove extension
-        fnames_out = [f[:f.rfind('.')] for f in sorted(os.listdir(dir_in))]
+        # fnames_out = [f[:f.rfind('.')] for f in sorted(os.listdir(dir_in))]
         # fnames_out = [os.path.basename(f).split('.')[0] for f in sorted(os.listdir(dir_in))]
 
+        fnames_out = sorted(os.listdir(dir_in))
+
         # add folder path
         fnames_out = utils.abs_path(dir_out,fnames_out)
-        
 
         for i, img_path in enumerate(fnames_in):
             print("running prediction for image: ", img_path)
             pred = self.run_prediction_single(img_path=img_path, return_logit=return_logit)
-            print("Saving images in", fnames_out[i]+".tif")
-            imsave(fnames_out[i]+".tif", pred)
-    
+
+            print("Saving images in", fnames_out[i])
+            _,spacing,origin,direction = utils.adaptive_imread(img_path, return_origin=True, return_direction=True)
+            adaptive_imsave(fnames_out[i], pred, spacing, origin, direction)
+                
     def load_train(self, 
         path, 
         load_best=False): # whether to load the best model
+        """Load a builder from a folder. The folder should have been created by the `self.build_train` method.
+        Can be use to restart a training. 
+
+        Parameters
+        ----------
+        path : str
+            Path of the log folder.
+        load_best : bool, default=False
+            Whether to load the best model or the final model.
         """
-        load a builder from a folder.
-        the folder should have been created by the self.run_training method.
-        """
+
+        # define config
+        self.config = utils.load_yaml_config(os.path.join(path,"log","config.yaml"))
 
         # setup the different paths from the folder
         self.base_dir = path
@@ -557,16 +738,11 @@ class Builder:
         print(self.model.load_state_dict(ckpt['model'], strict=False))
         if 'loss' in ckpt.keys(): self.loss_fn.load_state_dict(ckpt['loss'])
 
-        if not 'LR_START' in self.config.keys() or self.config.LR_START is None:
-            self.optim.load_state_dict(ckpt['opt'])
+        # if not 'LR_START' in self.config.keys() or self.config.LR_START is None:
+        self.optim.load_state_dict(ckpt['opt'])
 
-        if 'epoch' in list(ckpt.keys()): # tmp
-            self.initial_epoch=ckpt['epoch'] # definitive version 
-        # else: # tmp
-        #     with open(self.log_path, "r") as file:
-        #         last_line = file.readlines()[-1] # read the last line
-        #     last_line = last_line.split(',')
-        #     self.initial_epoch=int(last_line[0])
+        if 'epoch' in list(ckpt.keys()): 
+            self.initial_epoch=ckpt['epoch']+1 # definitive version 
         print('Restart training at epoch {}'.format(self.initial_epoch))
 
         self.build_dataset()
@@ -587,28 +763,69 @@ class Builder:
     def load_test(self, 
         path, 
         load_best=True): # whether to load the best model
+        """Load a builder from a folder. The folder should have been created by the `self.build_train` method.
+        Can be used to test the model on unseen data.
+
+        Parameters
+        ----------
+        path : str
+            Path of the log folder.
+        load_best : bool, default=True
+            Whether to load the best model or the final model.
         """
-        load a builder from a folder.
-        the folder should have been created by the self.run_training method.
-        """
+        # if the path is a list of path then multi-model mode
+        if type(path)==list:
+            print("We found a list of path for your model loading. Let's switch to multi-model mode!")
+            self.config = []
+            self.model = []
+            for p in path:
+                # define config
+                config = utils.load_yaml_config(os.path.join(p,"log","config.yaml"))
+                self.config += [config]
 
-        # setup the different paths from the folder
-        self.model_dir = os.path.join(path, 'model')
-        self.model_path = os.path.join(self.model_dir, self.config.DESC)
+                # setup the different paths from the folder
+                model_dir = os.path.join(p, 'model')
+                model_name = config.DESC + '_best.pth' if load_best else config.DESC + '.pth'
+                model_path = os.path.join(model_dir, model_name)
 
-        # call the build method
-        self.build_model(training=False)
+                # create the model
+                self.model += [read_config(config.MODEL, register.models)]
+                
+                print("Loading model weights for model:", self.model[-1])
 
-        # load the model and the optimizer
-        model_name_full = self.config.DESC + '_best.pth' if load_best else self.config.DESC + '.pth'
-        device = 'cuda' if torch.cuda.is_available() else 'cpu'
-        ckpt = torch.load(os.path.join(self.model_dir, model_name_full), map_location=torch.device(device))
-        print("Loading model from", os.path.join(self.model_dir, model_name_full))
+                # load the model
+                if getattr(self.model[-1], "load", None) is not None: 
+                    self.model[-1].load(model_path)
+                else:
+                    raise RuntimeError("For multi-model loading, please define a `load` method in your nn.Module definition.")
+        else:
+            # define config
+            self.config = utils.load_yaml_config(os.path.join(path,"log","config.yaml"))
 
-        # remove `module.` prefix
-        state_dict = {k.replace("module.", ""): v for k, v in ckpt['model'].items()}  
-        # remove `backbone.` prefix induced by multicrop wrapper
-        state_dict = {k.replace("backbone.", ""): v for k, v in state_dict.items()}
-        print(self.model.load_state_dict(state_dict, strict=False))
+            # setup the different paths from the folder
+            self.model_dir = os.path.join(path, 'model')
+            model_name = self.config.DESC + '_best.pth' if load_best else self.config.DESC + '.pth'
+            model_path = os.path.join(self.model_dir, model_name)
+
+            # create the model
+            self.model = read_config(self.config.MODEL, register.models) 
+
+            print(self.model)
+
+            # if the model has a 'load' method we use it
+            if getattr(self.model, "load", None) is not None: 
+                self.model.load(model_path)
+            # else try to load it here... but not a good practice, might be removed in the future 
+            # we keep this option to avoid forcing the definition of a 'load' method in the model.
+            else: 
+                device = 'cuda' if torch.cuda.is_available() else 'cpu'
+                ckpt = torch.load(model_path, map_location=torch.device(device))
+                print("Loading model from", model_path)
+
+                # remove `module.` prefix
+                state_dict = {k.replace("module.", ""): v for k, v in ckpt['model'].items()}  
+                # remove `backbone.` prefix induced by multicrop wrapper
+                state_dict = {k.replace("backbone.", ""): v for k, v in state_dict.items()}
+                print(self.model.load_state_dict(state_dict, strict=False))
 
 #---------------------------------------------------------------------------
\ No newline at end of file
