@@ -7,33 +7,26 @@ import argparse
 import os
 import numpy as np
 # from telegram_send import send
-from importlib import import_module
 
 from biom3d.builder import Builder
-from biom3d.utils import load_config, abs_listdir, versus_one, dice
+from biom3d.utils import abs_listdir, versus_one, dice, load_python_config
 
 
 #---------------------------------------------------------------------------
 # utils
 
-def train(config=None, log=None): 
-    cfg = None if not config else import_module(config).CONFIG
-    builder = Builder(config=cfg,path=log)
-    builder.run_training()
-    print("Training done!")
-
-def train_yaml(config=None, log=None): 
-    cfg = None if not config else load_config(config)
-    builder = Builder(config=cfg,path=log)
+def train(config=None, path=None): 
+    builder = Builder(config=config, path=path)
     builder.run_training()
     print("Training done!")
+    return builder
 
 #---------------------------------------------------------------------------
 # main unet segmentation
 
 def main_seg_pred_eval(
-    config=None,
-    log=None,
+    config_path=None,
+    path=None,
     dir_in=None,
     dir_out=None,
     dir_lab=None,
@@ -47,8 +40,7 @@ def main_seg_pred_eval(
     """
     # train
     print("Start training")
-    cfg = import_module(config).CONFIG
-    builder_train = Builder(config=cfg,path=log)
+    builder_train = Builder(config=config_path,path=path)
     if freeze_encoder:
         builder_train.model.freeze_encoder()
     builder_train.run_training()
@@ -62,8 +54,8 @@ def main_seg_pred_eval(
     if dir_in is not None and dir_out is not None:
         print("Start inference")
         builder_pred = Builder(
-            config=cfg,
-            path=train_base_dir, 
+            config=config_path,
+            path=path,
             training=False)
 
         dir_out = os.path.join(dir_out,os.path.split(train_base_dir)[-1]) # name the prediction folder with the model folder name
@@ -85,9 +77,9 @@ def main_seg_pred_eval(
                     fct=dice, 
                     in_path=list_abs[1][idx], 
                     tg_path=list_abs[0][idx], 
-                    num_classes=cfg.NUM_CLASSES if cfg.USE_SOFTMAX else (cfg.NUM_CLASSES+1), 
+                    num_classes=(builder_pred.config.NUM_CLASSES+1), 
                     single_class=None)]
-                print("Metric result:", print(results[-1]))
+                print("Metric result:", results[-1])
             print("Evaluation done! Average result:", np.mean(results))
             # send(messages=["Evaluation done of model {}! Average result: {}".format(dir_out, np.mean(results))])
         
@@ -118,15 +110,14 @@ def main_pretrain_seg_pred_eval(
     """
     # pretraining
     print("Start pretraining")
-    cfg = import_module(pretrain_config).CONFIG
-    builder = Builder(config=cfg,path=log)
+    builder = Builder(config=pretrain_config,path=log)
     if path_encoder is None:
         builder.run_training()
     print("Pretraining is done!")
 
     # train
     print("Start training")
-    cfg = import_module(train_config).CONFIG
+    cfg = load_python_config(train_config)
     # cfg.MODEL.kwargs.encoder_ckpt = os.path.join(builder.model_dir,builder.config.DESC+"_best.pth")
 
     if path_encoder and not model_encoder:
@@ -176,7 +167,7 @@ def main_pretrain_seg_pred_eval(
                     tg_path=list_abs[0][idx], 
                     num_classes=cfg.NUM_CLASSES if cfg.USE_SOFTMAX else (cfg.NUM_CLASSES+1), 
                     single_class=None)]
-                print("Metric result:", print(results[-1]))
+                print("Metric result:", results[-1])
             print("Evaluation done! Average result:", np.mean(results))
             # send(messages=["Evaluation done of model {}! Average result: {}".format(dir_out, np.mean(results))])
 
@@ -215,12 +206,10 @@ if __name__=='__main__':
     parser = argparse.ArgumentParser(description="Main training file.")
     parser.add_argument("-n", "--name", type=str, default="train",
         help="Name of the tested method. Valid names: {}".format(valid_names.keys()))
-    parser.add_argument("-c", "--config", type=str, default="configs.unet_pancreas",
-        help="Name of the python configuration file (full path and without the .py)")
-    parser.add_argument("-pc", "--pretrain_config", type=str, default="configs.vgg-dino_pancreas",
-        help="Name of the python configuration file for the pretraining (full path and without the .py)")
-    parser.add_argument("--config_yaml", type=str, default=None,
-        help="Name of the configuration file stored in yaml format (full path and without the .py)")
+    parser.add_argument("-c", "--config", type=str, default=None,
+        help="Name of the python or yaml configuration file")
+    parser.add_argument("-pc", "--pretrain_config", type=str, default=None,
+        help="Name of the python or yaml configuration file for the pretraining")
     parser.add_argument("-l", "--log", type=str, default=None,
         help="Name of the log folder. Used to resume model training")
     parser.add_argument("-le","--path_encoder", type=str, default=None,
@@ -229,11 +218,11 @@ if __name__=='__main__':
         help="Whether the encoder is a model encoder or a simple encoder.") 
     parser.add_argument("-fr", "--freeze_encoder", default=False,  action='store_true', dest='freeze_encoder',
         help="Whether to freeze or not the encoder.") 
-    parser.add_argument("-i", "--dir_in", type=str, default="/home/gumougeot/all/codes/python/3dnucleus/data/pancreas/imagesTs_small",
+    parser.add_argument("-i", "--dir_in", type=str, default=None,
         help="Path to the input image directory")
-    parser.add_argument("-o", "--dir_out", type=str, default="/mnt/52547A99547A8011/data/preds",
+    parser.add_argument("-o", "--dir_out", type=str, default=None,
         help="Path to the output prediction directory")  
-    parser.add_argument("-a", "--dir_lab", type=str, default="/home/gumougeot/all/codes/python/3dnucleus/data/pancreas/labelsTs_small",
+    parser.add_argument("-a", "--dir_lab", type=str, default=None,
         help="Path to the input image directory")  
     args = parser.parse_args()
 
@@ -259,9 +248,7 @@ if __name__=='__main__':
             dir_out=args.dir_out,
             dir_lab=args.dir_lab,
             )
-    elif args.config_yaml is not None:
-        train_yaml(config=args.config_yaml, log=args.log)
     else:
-        train(config=args.config, log=args.log)
+        train(config=args.config, path=args.log)
 
 #---------------------------------------------------------------------------
